#include <fcntl.h>
#include <linux/fd.h>
#include <stdio.h>
#include "sq80.h"

#define DISK_DUMP	0
#define DISK_COPY	1
#define DISK_TO_DISK	2
#define DISK_TO_FILE	3
#define FILE_COPY	4
#define FILE_TO_DISK	5
#define FILE_TO_FILE	6

void usage(void){
        printf("sq80copy drive -if=file -of=file [-nX] [-pY]\n");
	printf("\tDrive is\n");
	printf("\ta:\tcomplete disk (dump)\n");
	printf("\tb:\tprogram banks\n");
	printf("\tp:\tsingle programs\n");
	printf("\ts:\tsequencer files\n");
        printf("\tParameters:\n");
        printf("\t-if\tinput file\n");
        printf("\t-of\toutput file\n");
	printf("\tX\tfile number (not used for dump)\n");
	printf("\tOptional Parameter:\n");
	printf("\tY\tposition in file (sound bank only)\n");
        exit(0);
}

void error(const char *message){
        printf("%s\n",message);
        exit(-1);
}

int translate_chs(int track,int head, int sector) {
int ptr=0;
        if (track>0) ptr=track*11264;                           /* complete track offset */
        if (head==1) ptr+=5632;                                 /* "half" track offset */
        if ((sector>0) && (sector!=5)) ptr+=sector*1024;        /* big sector offset */
        if (sector==5) ptr+=5120;                               /* small sector offset */
        return ptr+10;                                          /* header size offset */
}

geo_t *get_prog_pos(geo_t *prgpos, int pnum) {
	switch( ((pnum-1)&64)|(((pnum-1)&63)+2) ) {
        	case 0x06:      prgpos->st=0x42; prgpos->sh=0; break;
                case 0x19:      prgpos->st=0x42; prgpos->sh=1; break;
                case 0x1f:      prgpos->st=0x43; prgpos->sh=1; break;
                case 0x39:      prgpos->st=0x44; prgpos->sh=0; break;
                case 0x3f:      prgpos->st=0x44; prgpos->sh=1; break;
                case 0x53:      prgpos->st=0x45; prgpos->sh=1; break;
                case 0x6c:      prgpos->st=0x46; prgpos->sh=0; break;
                case 0x73:      prgpos->st=0x46; prgpos->sh=1; break;
                default:        prgpos->st=((pnum-1)&63)+2;
                                prgpos->sh=((pnum-1)&64)>>6;
                                break;
	}
        prgpos->ss=5; 
	prgpos->et=prgpos->st; prgpos->eh=prgpos->sh; prgpos->es=prgpos->ss;
	return prgpos;
}

geo_t *get_file_pos(geo_t *filepos, int pnum, int ftype) {
int fnum;
	if (ftype!=PRG) {
		switch(ftype) {
			case BNK:	fnum=pnum+9;
					break;
			case SEQ:
			case SNG:
			case SYX:
			case ASX:	fnum=pnum-1;
					break;
		}	
		filepos->st=geo[fnum].st; filepos->sh=geo[fnum].sh; filepos->ss=geo[fnum].ss;
		filepos->et=geo[fnum].et; filepos->eh=geo[fnum].eh; filepos->es=geo[fnum].es;
	} else return get_prog_pos(filepos,pnum);

	return filepos;
}

void sqseek(int dnum, int c) {
char dummy[256];
	sprintf(dummy,"fdrawcmd seek %i %i 2>/dev/null",dnum&3,c);
	system(dummy);
}

void sqtune(int dnum) {
int i; 
char dummy[256];
        /* reset floppy */
        if ((i=open(floppy[dnum&3],O_NDELAY))<0) error("Cant access floppy drive.\n");
        if (ioctl(i,FDRESET,2)<0) {
       	        close(i);
               	error("Cant reset floppy drive.");
        }
       	close(i);

        /* tune to SQ80 disk */
       	sprintf(dummy,"fdrawcmd readid %i rate=2 2>/dev/null",dnum&3);
        system(dummy);
       	sprintf(dummy,"fdrawcmd recalibrate %i 2>/dev/null",dnum&3);
       	system(dummy);
}

int sizecode(int s) {
	if (s==5) return 2; else return 3;
}

int sectlen(int s) {
	if (s==5) return 512; else return 1024;
}

void checkdisk(int disk) {
int i;
	if ( (i=open(floppy[disk],O_NDELAY|O_RDONLY))<0 ) {
		printf("Cant access drive %s.\n",floppy[disk]);
		exit(-1);
	}
	close(i);
	sqtune(disk);
}

char *readheader(char *fname, char *dummy) {
FILE *fd;
	if ( (fd=fopen(fname,"r"))==NULL ) error("Cant read file.\n");
	if (fread(dummy,1,10,fd)!=10) {
		fclose(fd);
		error("Short header.\n");
	}
	dummy[10]=0;
	fclose(fd);
}

void checkseqheader(char *fname) {
char dummy[256];
	readheader(fname,dummy);
	if ( (strcmp(dummy,"!SQ80SEQU!")!=0) && (strcmp(dummy,"!SQ80SONG!")!=0) && (strcmp(dummy,"!SQ80SYSX!")!=0) )
		error("Wrong header.\n");
}

void checkheader(char *fname, char *comp) {
char dummy[256];
	readheader(fname,dummy);
	if (strcmp(dummy,comp)!=0) error("Wrong file.\n");
}

void checkfile(char *fname, int ftype) {
	switch(ftype) {
		case ASX:	checkseqheader(fname);
				break;
		default:	checkheader(fname,ft_hdr[ftype]);
				break;
	}
}
	
void checksize(int ftype, int fnum, int pnum) {
	switch(ftype) {
		case BNK:	if ((fnum<1) && (fnum>40)) error("Bank file range from 1 to 40.\n");
				if ((pnum<1) && (pnum>40)) error("Program numbers inside bank files range from 1 to 40.\n");
				break;
		case PRG:	if ((fnum<1) && (fnum>128)) error("Program files range from 1 to 128.\n");
				break;
		case ASX:	if ((fnum<1) && (fnum>10)) error("Sequencer files range from 1 to 10.\n");
				break;
	}
}

void write_header(char *outfile, int ftype, int fnum) {
char dummy[64];
FILE *fd;
	sprintf(dummy,"%s%c%c%c%s",ft_hdr[ftype],diskdir[fnum].type,(diskdir[fnum].size&0xffff)>>8,diskdir[fnum].size&0xff,diskdir[fnum].name);
	if ( (fd=fopen(outfile,"w"))==NULL) error("Can't write to file.\n");
	fwrite(dummy,1,23,fd);
	fclose(fd);	
}

int getlength(int ftype) {
	if 	(ftype==PRG) return 1;
	else if (ftype==BNK) return 4;
	else return 64;
}

void read_message(int c, int h, int s) {
	printf("Reading track %2i, head %i, sector %i (%4i bytes).\r",c,h,s,sectlen(s));
	fflush(stdout);
}

void write_message(int c, int h, int s) {
	printf("Writing track %2i, head %i, sector %i (%4i bytes).\r",c,h,s,sectlen(s));
	fflush(stdout);
}

char *process_dirbuf() {
int i;
	for (i=0;i<178;i++) {
                if (i<50) {                             // sequencer & bank files
                        diskdir[i].type=dirbuf[i*13];
                        diskdir[i].size=((dirbuf[i*13+11]&0xff)<<8)|(dirbuf[i*13+12]&0xff);     
                        memcpy(diskdir[i].name,dirbuf+i*13+1,10);
                        diskdir[i].name[10]=0;
                } else {                                // single programs
                        diskdir[i].type=PRG;
                        diskdir[i].size=102;
                        memcpy(diskdir[i].name,dirbuf+650+(i-50)*6,6);
                        diskdir[i].name[6]=0;
                }
        }
}

char *get_disk_dir(int disk) {
char dummy[256];
FILE *fd;
int i;
	for (i=0;i<4;i++) {
		if ((i%2)==0) sqseek(disk,dirpos[i].c);
		sprintf(dummy,"fdrawcmd read %i %i %i 5 2 0 0x1b 0xff length=512 rate=2 2>/dev/null",(disk&3)|(dirpos[i].h<<2),dirpos[i].c,dirpos[i].h);
		fd=popen(dummy,"r");
			fread(dirbuf+i*512,1,512,fd);
		fclose(fd);
	}

	process_dirbuf();
}

char *get_dump_dir(char *fname) {
int i,ptr;
char dummy[256];
FILE *fd;
	for (i=0;i<4;i++) {
		ptr=translate_chs(dirpos[i].c,dirpos[i].h,dirpos[i].s);
		fd=fopen(fname,"r");
			if (fseek(fd,ptr,SEEK_SET)<0) {
				fclose(fd);
				error("Short file.\n");
			}
			if (fread(dirbuf+i*512,1,512,fd)!=512) {;
				fclose(fd);
				error("Short file.\n");
			}
		fclose(fd);
	}

	process_dirbuf();
}

char *write_disk_dir(int disk) {
int i;
char dummy[256];
FILE *fd;	

	for (i=0;i<50;i++) {
		dirbuf[i*13]=diskdir[i].type&0xff;
		memcpy(dirbuf+1+i*13,&diskdir[i].name,10);
		dirbuf[i*13+11]=(diskdir[i].size&0xffff)>>8;
		dirbuf[i*13+12]=diskdir[i].size&0xff;
	}	

	for (i=0;i<128;i++) 
		memcpy(dirbuf+i*6+650,&diskdir[i+50].name,6);
	
	memcpy(dirbuf+2046,"SQ",2);

	for (i=0;i<4;i++) {	
		if ((i%2)==0) sqseek(disk,dirpos[i].c);
		sprintf(dummy,"fdrawcmd write %i %i %i 5 2 0 0x1b 0xff length=512 rate=2 2>/dev/null",(disk&3)|(dirpos[i].h<<2),dirpos[i].c,dirpos[i].h);
		fd=popen(dummy,"w");
			fwrite(dirbuf+i*512,1,512,fd);
		pclose(fd);
	}
}

int get_type(int fnum, int ftype) {
	if (ftype==ASX) return diskdir[fnum-1].type;
	else return ftype;
}

int get_dirpos(int fnum, int ftype) {
	if (ftype==BNK) return fnum+9;
	else if (ftype==PRG) return fnum+49;
	else return fnum-1;
}

/*******************/
/* copy SQ80 disks */
/*******************/
void disk_copy(int indisk, int outdisk) {
char cmd[256], buffer[901120];
int c,h,s,sc,sl,offs;
FILE *fd;
	checkdisk(indisk);
	if (indisk!=outdisk) checkdisk(outdisk);

	sqseek(indisk,0); 
	c=0; h=0; s=0; offs=0;
	printf("Reading contents of disk %s.\n",floppy[indisk]);
	while ( !((c==80) && (h==0) && (s==0)) ) {
		sc=sizecode(s); sl=sectlen(s);
		read_message(c,h,s);
		sprintf(cmd,"fdrawcmd read %i %i %i %i %i 0 0x1b 0xff length=%i rate=2 2>/dev/null",(indisk&3)|(h<<2),c,h,s,sc,sl);
		fd=popen(cmd,"r");
			fread(buffer+offs,1,sl,fd);
		pclose(fd);
		offs+=sl;
		if (s<5) s++;
		else {
			s=0;
			if (h<1) h++;
			else {
				h=0;
				c++;
				sqseek(indisk,c);			
			}
		}
	}

	printf("\n\nPlease enter destination disk and press return key.\n(Ensure that write protection is off.)\n");
	getchar();

	checkdisk(outdisk);

	c=0; h=0; s=0; offs=0;
	sqseek(outdisk,0);
	printf("Writing contents from disk %s to disk %s.\n",floppy[indisk],floppy[outdisk]);
	while ( !( (c==80) && (h==0) && (s==0) ) ) {
		sc=sizecode(s); sl=sectlen(s);
		write_message(c,h,s);
		sprintf(cmd,"fdrawcmd write %i %i %i %i %i 0 0x1b 0xff length=%i rate=2 2>/dev/null",(indisk&3)|(h<<2),c,h,s,sc,sl);
		fd=popen(cmd,"w");
			fwrite(buffer+offs,1,sl,fd);
		pclose(fd);
		offs+=sl;

		if (s<5) s++;
		else {
			s=0;
			if (h<1) h++;
			else {
				h=0;
				c++;
				sqseek(outdisk,c);			
			}
		}
	}
}

/**********************/
/* copy out SQ80 file */
/**********************/
void dfcopy(int indisk,char *outfile,int fnum,int ftype,int pnum){
geo_t myfile;
int c,h,s,i,len,sc,sl,rtype,rnum,h_changed;
char cmd[256], buffer[4096];
FILE *fd;
	checksize(ftype,fnum,pnum);
	checkdisk(indisk);

	get_disk_dir(indisk);
	get_file_pos(&myfile,fnum,ftype);
	
	rtype=get_type(fnum,ftype);
	rnum =get_dirpos(fnum,ftype);

	if ( (ftype!=BNK)||((ftype==BNK)&&(pnum==-1)) ) 
		write_header(outfile,rtype,rnum);	

	len=getlength(ftype);	
	c=myfile.st; h=myfile.sh&0x01; s=myfile.ss; h_changed=(myfile.sh&0x10)>>4;

	if ( (ftype!=BNK)||((ftype==BNK)&&(pnum==-1)) ) 
		printf("Copying out file %s from disk %s.\n",diskdir[rnum].name,floppy[indisk]);
	else
		printf("Copying out program %i from bank %s of disk %s.\n",pnum,diskdir[rnum].name,floppy[indisk]);
	fflush(stdout);

#ifdef SLOW_FLOPPY
	for (i=0;i<=c;i++) sqseek(indisk,i);
#else
	sqseek(indisk,c);
#endif

	for(i=0;i<len;i++) {
		sc=sizecode(s); sl=sectlen(s);
		read_message(c,h,s);
		if ( (ftype!=BNK)||((ftype==BNK)&&(pnum==-1)) ) {
			sprintf(cmd,"fdrawcmd read %i %i %i %i %i 0 0x1b 0xff length=%i rate=2 2>/dev/null >> %s",(indisk&3)|(h<<2),c,h,s,sc,sl,outfile);
			system(cmd);
		} else {
			sprintf(cmd,"fdrawcmd read %i %i %i %i %i 0 0x1b 0xff length=%i rate=2 2>/dev/null",(indisk&3)|(h<<2),c,h,s,sc,sl);
			fd=popen(cmd,"r");
			fread(buffer+i*1024,1,1024,fd);
			pclose(fd);
		}
		if (s<4) s++;
		else {
			s=0;

			if ((ftype==ASX)&&(h_changed==1)) {	// sequencer files
				c++;
				h_changed=0;
			} else if ((ftype==ASX)&&(h_changed==0)) {
				if (h==0) h=1; else h=0;
				h_changed=1;
			} else 					// bank & program files
				c++;
			sqseek(indisk,c);
		}
	}
				
	if (ftype==ASX) {	// get small sector of sequencer file
		read_message(c,h,5);
		sprintf(cmd,"fdrawcmd read %i %i %i 5 2 0 0x1b 0xff length=512 rate=2 2>/dev/null >> %s",(indisk&3)|(h<<2),c,h,outfile);
		system(cmd);
	}		

	/* copy out single program from bank file */
	if ((ftype==BNK)&&(pnum!=-1)) {
		fd=fopen(outfile,"w");
		sprintf(cmd,"%s%c%c%c",ft_hdr[PRG],PRG,0,102);
		fwrite(cmd,1,13,fd);
		printf("offset: %i (%i)\n",(pnum-1)*102,pnum);
		fwrite(buffer+(pnum-1)*102,1,102,fd);
		fclose(fd);
	}		
}

/*********************/
/* copy in SQ80 file */
/*********************/
void fdcopy(char *infile,int outdisk,int fnum,int ftype,int pnum){
geo_t myfile;
char buffer[66048], dummy[256];
int rnum, len, c,h,s,sc,sl, i, h_changed;
FILE *fd;
	checksize(ftype,fnum,pnum);
	checkdisk(outdisk);

	get_disk_dir(outdisk);

	get_file_pos(&myfile,fnum,ftype);
	
	len=getlength(ftype);	
	c=myfile.st; h=myfile.sh&0x01; s=myfile.ss; h_changed=(myfile.sh&0x10)>>4;

#ifdef SLOW_FLOPPY
	for(i=0;i<=c;i++) sqseek(outdisk,i);
#else
	sqseek(outdisk,c);
#endif	
	rnum =get_dirpos(fnum,ftype);
	if ( (ftype!=BNK)||((ftype==BNK)&&(pnum==-1)) ) 
		printf("Copying in file %s to disk %s.\n",infile,floppy[outdisk]);
	else
		printf("Copying in program %i to bank %s of disk %s.\n",pnum,diskdir[rnum].name,floppy[outdisk]);
	fflush(stdout);

	/* import single program into bank file */
	if ((ftype==BNK)&&(pnum!=-1)) {
		checkfile(infile,PRG);

		/* read in bank */
		for(i=0;i<4;i++) {
			read_message(c,h,s);
			sprintf(dummy,"fdrawcmd read %i %i %i %i 3 0 0x1b 0xff length=1024 rate=2 2>/dev/null",(outdisk&3)|(h<<2),c,h,s);
			fd=popen(dummy,"r");
			fread(buffer+i*1024,1,1024,fd);
			pclose(fd);
			if (s<4) s++;
			else {
				s=0;
				c++;
				sqseek(outdisk,c);
			}
		}

		/* modify sound */
		fd=fopen(infile,"r");
		fseek(fd,23,SEEK_SET);
		fread(buffer+(pnum-1)*102,1,102,fd);
		fclose(fd);

		/* write back bank */
		c=myfile.st; h=myfile.sh; s=myfile.ss;
//#ifdef SLOW_FLOPPY
//		for(i=0;i<=c;i++) sqseek(outdisk,i);
//#else
		sqseek(outdisk,c);
//#endif	
		for(i=0;i<4;i++) {
			write_message(c,h,s);
			sprintf(dummy,"fdrawcmd write %i %i %i %i 3 0 0x1b 0xff length=1024 rate=2 2>/dev/null",(outdisk&3)|(h<<2),c,h,s);
			fd=popen(dummy,"w");
			fwrite(buffer+i*1024,1,1024,fd);
			pclose(fd);
			if (s<4) s++;
			else {
				s=0;
				c++;
				sqseek(outdisk,c);
			}
		}
		
	/* import file */
	} else {
		checkfile(infile,ftype);

		fd=fopen(infile,"r");
		fseek(fd,23,SEEK_SET);				// skip header, type & name
		if 	(ftype==BNK) fread(buffer,1,4080,fd);	// get data
		else if (ftype==PRG) fread(buffer,1,102,fd);	
		else fread(buffer,1,66048,fd);	
		fclose(fd);
		
		for(i=0;i<len;i++) {
			write_message(c,h,s);
			sc=sizecode(s); sl=sectlen(s);
			sprintf(dummy,"fdrawcmd write %i %i %i %i %i 0 0x1b 0xff length=%i rate=2 2>/dev/null",(outdisk&3)|(h<<2),c,h,s,sc,sl);
			fd=popen(dummy,"w");
				fwrite(buffer+i*1024,1,1024,fd);	
			pclose(fd);
			if (s<4) s++;
			else {
				s=0;	

				if ((ftype==ASX)&&(h_changed==1)) {	// sequencer files
					c++;
					h_changed=0;
				} else if ((ftype==ASX)&&(h_changed==0)) {
					if (h==0) h=1; else h=0;
					h_changed=1;
				} else 					// bank & program files
					c++;
				sqseek(outdisk,c);

			 }
		}				

		if (ftype==ASX) {	// write small sector of sequencer file
			write_message(c,h,5);
			sprintf(dummy,"fdrawcmd write %i %i %i 5 2 0 0x1b 0xff length=512 rate=2 2>/dev/null",(outdisk&3)|(h<<2),c,h);
			fd=popen(dummy,"w");
			fwrite(65536+buffer,1,512,fd);
			pclose(fd);
		}		

		sqseek(outdisk,0);

		printf("\nUpdating directory.\n");
		fd=fopen(infile,"r");
			fseek(fd,10,SEEK_SET);
			fread(dummy,1,1,fd);
			diskdir[rnum].type=dummy[0]&0xff;
			fread(dummy,1,2,fd);
			diskdir[rnum].size=((dummy[0]&0xff)<<8)|(dummy[1]&0xff);
			fread(diskdir[rnum].name,1,10,fd);
		fclose(fd);
		write_disk_dir(outdisk);						
	}
}
		
int main(int argc, char **argv) {
char infile[256], outfile[256], dummy[256];
int dir=-1, fnum=-1, pnum=-1, indisk=-1, outdisk=-1,ftype,i,mode;

        argv++; argc--;

        if ((argc>5)||(argc<3)) usage();
        
        while(*argv!=NULL){
                if 	(strncmp(*argv,"-if=",4)==0) strcpy(infile,*argv+4);
                else if (strncmp(*argv,"-of=",4)==0) strcpy(outfile,*argv+4);
		else if (strncmp(*argv,"-n",2)==0) fnum=strtoul(*argv+2,0,0);
		else if (strncmp(*argv,"-p",2)==0) pnum=strtoul(*argv+2,0,0);
		else if (strcmp(*argv,"a:")==0) ftype=ALL;
		else if (strcmp(*argv,"b:")==0) ftype=BNK;
		else if (strcmp(*argv,"s:")==0) ftype=ASX;
		else if (strcmp(*argv,"p:")==0) ftype=PRG;
                else usage();
                argv++;
        }

        for(i=0;i<4;i++) {
                if (strncmp(infile,floppy[i],8)==0) {
                        indisk=i;
                }
                if (strncmp(outfile,floppy[i],8)==0) {
                        outdisk=i;
                }
        }

	if (fnum==-1) {
		if 	( (indisk>=0) && (outdisk>=0) ) mode=DISK_COPY;
		else if ( (indisk<0)  && (outdisk<0) )	mode=FILE_COPY;
		else if ( (indisk>=0) && (outdisk<0) )	mode=DISK_DUMP;
		else usage();
	} else {
		if 	( (indisk>=0) && (outdisk>=0) ) mode=DISK_TO_DISK;
		else if ( (indisk>=0) && (outdisk<0) )	mode=DISK_TO_FILE;
		else if ( (indisk<0)  && (outdisk>=0) )	mode=FILE_TO_DISK;
		else if ( (indisk<0)  && (outdisk<0) )	mode=FILE_TO_FILE;
		else usage();
	}

/* 
	// doesn't help - so I need to keep that noisy single step seek //
	if ((indisk>=0)||(outdisk>=0)) {
		sprintf(dummy,"fdrawcmd specify 0 0");	// slow floppy, use DMA
		system(dummy);
	}
*/

	switch(mode) {
		case DISK_DUMP:		sprintf(dummy,"sq80dump -if=%s -of=%s",infile,outfile);
					system(dummy);
					break;
		case DISK_COPY:		disk_copy(indisk,outdisk);
					break;
		case FILE_COPY:		sprintf(dummy,"cp %s %s",infile,outfile);
					system(dummy);
					break;
		case DISK_TO_FILE:	dfcopy(indisk,outfile,fnum,ftype,pnum);
					break;
		case FILE_TO_DISK:	fdcopy(infile,outdisk,fnum,ftype,pnum);
					break;
	}	
	printf("\n");
	if (indisk>0) sqseek(indisk,0);
	if ((outdisk>0)&&(outdisk!=indisk)) sqseek(outdisk,0);
	exit(0);
}
